package at.ac.uibk.gitsearch.service;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.gitlab4j.api.Constants.TokenType;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.AccessLevel;
import org.gitlab4j.api.models.Permissions;
import org.gitlab4j.api.models.Project;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.FileCopyUtils;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;

import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.SearchResultDTO;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
public class SearchService {

	private final Logger log = LoggerFactory.getLogger(SearchService.class);

	@Autowired
	protected MetaDataRepository metaDataRepository;
	@Autowired
	protected TokenProvider tokenProvider;

	/** just for test implementation **/
	public static int NUM_TESTRESULTS = 23;

	public SearchService() {

	}

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getKeywordsAutoComplete(String keyWordPrefix) throws IOException {
		return metaDataRepository.getKeywordsAutoComplete(keyWordPrefix);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getCreatorAutoComplete(String creatorPrefix) throws IOException {
		return metaDataRepository.getCreatorAutoComplete(creatorPrefix);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getContributorAutoComplete(String contributorPrefix) throws IOException {
		return metaDataRepository.getContributorAutoComplete(contributorPrefix);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param programmingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getProgrammingLanguageAutoComplete(String programmingLanguagePrefix) throws IOException {
		return metaDataRepository.getProgrammingLanguageAutoComplete(programmingLanguagePrefix);
	}

	/**
	 * currently just a dummy implementation. TODO: first, and length is redundant
	 * to searchInput.getPage()
	 *
	 * @param searchInput the query definition
	 * @param first       the index of the first record to be returned
	 * @param length      the number of records
	 */
	public SearchResultsDTO searchResultPage(SearchInputDTO searchInput, long first, int length) throws IOException {
		log.debug("Searchrequest for {} ", searchInput);
		final SearchResultsDTO pageDetails = metaDataRepository.pageDetails(searchInput, length);
		Optional<String> accessTokenO = tokenProvider.getGitLabAccessToken();
		boolean isPresent = accessTokenO.isPresent();
		GitLabApi gitLabApi = null;
		if(isPresent) {
			String idToken = accessTokenO.get();
			final Optional<String> gitLabAccessIssuer = tokenProvider.getGitLabAccessIssuer();
			if(! gitLabAccessIssuer.isPresent()) {
				log.warn("accessToken defined, but no gitlabAccess Issuer found?");
			} else {

	         gitLabApi = new GitLabApi(gitLabAccessIssuer.get(), TokenType.OAUTH2_ACCESS, idToken);

	        }
		}

		final GitLabApi gitLabApif = gitLabApi;
		pageDetails.setSearchResult(pageDetails.getSearchResult().stream().filter(p->testAccess(p, gitLabApif)).collect(Collectors.toList()));
		pageDetails.getSearchResult().stream().forEach(md -> fixImageURL(md, gitLabApif));

		return pageDetails;
//		return readTestResults(searchInput.getFulltextQuery(), first, length);
	}

	private boolean testAccess(SearchResultDTO p, GitLabApi gitLabApi) {
		if(true) return true;
		if(gitLabApi==null) return false;
					try {
						Project gitProject = gitLabApi.getProjectApi().getProject(p.getProject().getProject_id());
						final Permissions permissions = gitProject.getPermissions();
						log.info("Permissions {}", permissions);
						boolean access =
								permissions.getProjectAccess()!=null  &&
								permissions.getProjectAccess().getAccessLevel().toValue() >= AccessLevel.GUEST.toValue() ||
								permissions.getGroupAccess() !=null &&
								permissions.getGroupAccess().getAccessLevel().toValue() >= AccessLevel.GUEST.toValue() ;
						return access;
					} catch (GitLabApiException e) {
						log.debug("It seems that current user cannot access project {}({}) for {}", p.getTitle(), p.getIdentifier(), p.getProject().getProject_id());
						return false;
					}
	}

	private void fixImageURL(SearchResultDTO metaData, GitLabApi gitLabApi) {
		String image = metaData.getImage();

		URI url = null;
		try {
			if(image != null)	url = new URI(image);
		if(url!=null && url.isAbsolute()) return;

		final String baseUrl =
				ServletUriComponentsBuilder.fromCurrentContextPath().build().toUriString();
		final boolean isJava = metaData.getProgrammingLanguage() != null && Arrays.stream(metaData.getProgrammingLanguage()).anyMatch(s -> s.startsWith("JAVA"))
				&& metaData.hashCode() % 2 == 0; // nur um ein wenig Abwechslung zu bekommen :-)
		final boolean isLatexFormat = metaData.getFormat()!=null && Arrays.stream(metaData.getFormat()).anyMatch(s -> s.startsWith("latex"));
		final boolean isPython = metaData.getProgrammingLanguage() != null && Arrays.stream(metaData.getProgrammingLanguage()).anyMatch(s -> s.startsWith("python"));
		if(isJava) {
			metaData.setImage(baseUrl + "/content/img/java.png");
			return;
		}
		if(isLatexFormat) {
			metaData.setImage(baseUrl + "/content/img/latex.png");
			return;
		}
		if(isPython) {
			metaData.setImage(baseUrl + "/content/img/python.png");
			return;
		}
		metaData.setImage(baseUrl + "/content/img/gitLab.png");
		if(true) return;

		if(gitLabApi==null) return;

		try {
			Project gitProject = gitLabApi.getProjectApi().getProject(metaData.getProject().getProject_id());
			final String httpUrlToRepo = gitProject.getHttpUrlToRepo();
			String baseRepoURL = httpUrlToRepo.replaceFirst(".git", "/-/raw/master/");
			final URI resolvedImageUrl = new URI(baseRepoURL).resolve(url);
			metaData.setImage(resolvedImageUrl.toASCIIString());
		} catch (GitLabApiException e) {
			log.debug("It seems that current user cannot access project {}({}) for {}", metaData.getTitle(), metaData.getIdentifier(), metaData.getProject().getProject_id());
		}

		} catch (URISyntaxException e) {
			log.warn("Cannot parse image url {} for {}", metaData.getImage(), metaData.getGitURL());
		}
	}

	/**
	 * temporary static test data.
	 *
	 * @param infoString just a string to add to testdata description.
	 * @return
	 */
	private SearchResultsDTO readTestResults(String searchString, long first, int length) {
		ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
		mapper.findAndRegisterModules();

		List<SearchResultDTO> loadedMetaData = new ArrayList<SearchResultsDTO.SearchResultDTO>();

		for (String testFile : new String[] { "metaData1.yaml", "metaData2.yaml" }) {
			log.debug("reading test data from {} ", testFile);

			final File metaDataFile = new File("src/main/java/at/ac/uibk/gitsearch/service/testData/" + testFile);
			try {
				byte[] fileContent = FileCopyUtils.copyToByteArray(metaDataFile);
				SearchResultDTO searchResultDTO = mapper.readValue(fileContent, SearchResultDTO.class);
				loadedMetaData.add(searchResultDTO);
			} catch (IOException e) {
				log.error("Cannot read test search input from " + testFile, e);
			}
		}

		List<SearchResultDTO> result = new ArrayList<>();
		if (length > 0) {
			int metadataPos = 0;
			long recordNr = first;
			// This loop is only used for testing to get more results.
			for (int i = 0; i < 12; ++i) {
			    metadataPos = 0; // also for testing, remove later
                while (metadataPos < length && metadataPos < loadedMetaData.size()) {
                    final SearchResultDTO loadedMetaDataRecord = new SearchResultDTO(loadedMetaData.get(metadataPos++));
                    if (loadedMetaDataRecord.toString().contains(searchString)) {
                        loadedMetaDataRecord.setTitle("#" + (recordNr + 1) + " " + loadedMetaDataRecord.getTitle());
                        result.add(loadedMetaDataRecord);
                        recordNr++;
                        log.debug("Added MetaData Record {}", loadedMetaData);
                    }
                    log.debug("{} does not contain {}", loadedMetaData, searchString);
                }
            }

		}
		final SearchResultsDTO searchResultsDTO = new SearchResultsDTO(result, NUM_TESTRESULTS, first);

		return searchResultsDTO;

	}

}
